/*******************************************************************************
*                    tailor_c for QDOS/SMS2/SMSQ(/E) systems                   *
*                           and C68 compiler v4.20b                            *
*******************************************************************************/

#include <qdos.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "revision.h"

char _prog_name[]   = "gzip";
char _version[24]   = VERSION;
char _copyright[32] = __DATE__;
char *_endmsg = "\nPress a key to exit.";

long _stack = 32768L;

struct WINDOWDEF _condetails  = { 208, 1, 0, 7, 512, 256, 0, 0};

void (*_consetup)() = consetup_title;

int custom_expand (char * param, char ***argvptr, int *argcptr);

int (*_cmdwildcard)() = custom_expand;

/* custom cmdexpand */

#define FILEBUF_INIT    1024        /* Initial allocation size for buffer */
#define FILEBUF_INCR    1024        /* Increment size for buffer */

int custom_expand (char * param, char ***argvptr, int *argcptr)
{
    int     count,sl;
    size_t  bufsize;
    char    *filenamebuf;
    char    *ptr,*safeptr;

    /*
     *  Check to see if we should do wild card expansion.
     *  We only perform wildcard expansion if the parameter
     *  was not a string and if it contains one of the
     *  wild card characters.
     *
     *  We also do not expand any option that starts with '-'
     *  as we then assume that it is a unix stylew option.
     */
    if ((*param == '-') ||  (strpbrk(param,"*?") == NULL) ) {
        return 0;
    }

    if ((filenamebuf = malloc (bufsize = FILEBUF_INIT)) == NULL) {
        return -1;
    }
TRYAGAIN:
    count = getfnl(param, filenamebuf, bufsize, QDR_ALL);
    if (count == -1  && errno == ENOMEM) {
        /*
         *  We have overflowed the buffer, so we try
         *  to get a bigger buffer and try again.
         */
        bufsize += FILEBUF_INCR;
        if ((filenamebuf = realloc (filenamebuf, bufsize)) == NULL) {
            return -1;
        } else {
            goto TRYAGAIN;
        }
    }
    /*
     *  If no files were found, then return unexpanded.
     */
    if (count == 0) {
        free (filenamebuf);
        return 0;
    }
    /*
     *  Files were found, so add these to the list instead
     *  of the original parameter typed by the user.
     */
    for ( ptr=filenamebuf ; count > 0 ; count -- ) {
        *argvptr = (char **)realloc (*argvptr, (size_t)(((*argcptr) + 2) * sizeof (char *)));
        safeptr=(char *)malloc((size_t)(sl=strlen(ptr)+1));
        if (safeptr == NULL || *argvptr == NULL) {
            return -1;
        }
        (void) memcpy(safeptr,ptr,(size_t)sl);
        (*argvptr)[*argcptr] = safeptr;
        *argcptr += 1;
        ptr += sl;
    }
    free (filenamebuf);
    return *argcptr;
}

#ifdef DIRFIX   /* libc_a v4.20b got bugs in "opendir" and "readdir"... */

/*
 *          o p e n d i r
 *
 *  Unix compatible routione to open a directory
 *  in a fil;e system independent manner,.
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  21 Aug 93   DJW   - Removed assumption that open_qdir() sets up the
 *                      error codes when QDOS error code returned.
 *                    - Amended setting up of inode field to try and give
 *                      more chance of it being unique.
 *
 *  12 Jan 96   DJW   - Added check that underscore is always present in
 *                      the directory name used.
 *                    - Fixed problem with "." and "./" used as name.
 *                    - Improved algorithm for getting stored directory name.
 */

DIR * good_opendir(char * name)
{
    DIR *dirp;
    chanid_t ret;
    struct qdirect header;

    /*
     *  Start by checking the length of the
     *  name is within bounds
     */
    if(strlen(name) > MAXNAMELEN) {
        errno = ENAMETOOLONG;
        return NULL;
    }
    /*
     *  Get space for directory control block
     */
    if(!(dirp = (DIR *) malloc(sizeof(DIR)))) {
        return NULL;
    }
    dirp->dd_loc = 0L;
    /*
     *  Cope with opendir("."), opendir("._") or opendir("./")
     */
    if (strcmp(name, ".") && strcmp(name, "._") && strcmp(name, "./")) {
        (void) _mkname (dirp->dd_name, name);
    } else {
        (void) getcwd (dirp->dd_name, sizeof (dirp->dd_name));
    }
    /*
     *  Make sure that the trailing underscore is present
     */
    if( dirp->dd_name[strlen(dirp->dd_name)-1] != '_' ) {
        (void) strcat (dirp->dd_name,"_");
    }

    /*
     *  Try and access the name given as a directory
     *  Take error exit if access fails
     */
    if ((ret = open_qdir(dirp->dd_name)) <= 0) {
        if (ret< 0) {
            errno = EOSERR;
/*            _oserr = ret; */
        }
        return (NULL);
    }
    /*
     *  We want to try and ensure that the inode field is unique
     *  as a number of unix programs (particularily GNU ones)
     *  often compare two inodes to see if two files are really
     *  the same underlying file.
     *
     *  Use the file date as the inode, hopefully this gives
     *  a good chance of getting a unique value.
     */
    (void) fs_headr(ret, (timeout_t)-1, (char *)&header, sizeof(header));
    dirp->dd_buf.d_ino = header.d_update;
    dirp->dd_buf.d_off = (off_t)64;

    /*
     *  Close the channel id we got
     */
    (void) io_close(ret);
    return dirp;
}

/*
 *      r e a d d i r
 *
 *  Routine to return the next (UNIX style) directory entry.
 *
 *  NOTE.  It would be more efficient if we did not continually
 *     open/close the directory.  However, this is done so
 *     that the directory is not kept open across calls to
 *     this routine as this more closely emulates the way
 *     that one works in Unix.
 *
 *  Amendment History
 *  ~~~~~~~~~~~~~~~~~
 *  24 Apr 93   DJW   - It was possible to get out of this routine
 *                      without closing the directory channel if the
 *                      read_qdir() call failed.  This could result in
 *                      many channels being left open.
 *                      (problem reported by Erik Slagter)
 *                    - Use of 'goto' statements to common up exit paths
 *                      removed.  This is done automatically by the C68
 *                      peephole optimiser anyway.
 *                    - Changed fs_pos() calls to check return values
 *                      rather than _oserr to detect error conditions.
 *
 *  21 Aug 93   DJW   - Removed assumption that open_qdir() sets up the
 *                      error codes when QDOS error code returned.
 *
 *  09 Nov 93   DJW   - Slight change due to fact that d_namelen filed changed
 *                      to d_namlen to bring in line with UNIX standard
 *                      (reported by Phil Gaskell)
 *
 *  02 Aug 95   EAJ   - Adds underscore to the wildcard, and makes sure that
 *                      the underscore is included in "len".
 *                      This is so that the names returned, when the directory
 *                      supplied does not include the trailing underscore,
 *                      are not prefixed with an underscore !
 *                      (Once again, "diff" revealed this problem)
 *                      (code reworked by DJW to be more effecient).
 *
 *  12 Jan 96   DJW   - Changed to correctly add terminator after wild-card
 *                      (was adding it too late).
 *                    - Corrected error that chopped of first character of name.
 *                      (introduced when adding fix for no trailing underscore).
 *                      Problem and fix from Thierry Godefroy.
 *                    - Moved check for trailing underscore to opendir() where
 *                      it makes more sense for it to be present as it is then
 *                      only done once.
 */

struct dirent * good_readdir(DIR * dirp)
{
    chanid_t chid;
    struct qdirect qbuf;
    struct dirent *dp;
    char    devwc[MAXNAMELEN+2];
    size_t  len;
    long    reply;

    /*
     *  Check we have a valid pointer
     */
    if (dirp == NULL) {
        errno = EBADF;
        return NULL;
    }
    /*
     *  Open the QDOS directory
     */
    if((chid = open_qdir(dirp->dd_name)) <= 0) {
        if (chid < 0) {
/*            _oserr = chid; */
            errno = EOSERR;
        }
    return NULL;
    }

    strcpy(devwc, dirp->dd_name);
    len = strlen (devwc);
    devwc[len] = '*';               /* Add wildcard character */
    devwc[len+1] = '\0';            /* Add NULL terminator */

    /*
     *  Seek to the correct position
     */
    if ((reply=fs_pos(chid, dirp->dd_loc, (short)0)) < 0 ) {
/*        _oserr = reply; */
        errno = EOSERR;
        (void) io_close (chid);
        return NULL;
    }
    dp = &dirp->dd_buf;
    /*
     *  Read the next directory entry
     */
    if((reply=read_qdir(chid, devwc, dp->d_name, &qbuf, 0)) != 1) {
        (void) io_close(chid);
        return NULL;
    }
    /*
     *  Save the new position
     */
    dirp->dd_loc = fs_pos(chid,0L,1L);

    /*
     *  Remove the un-needed parts of the name
     *  (the parts common to dirp->dd_name
     */
    (void) strcpy( dp->d_name, &dp->d_name[len]);
    dp->d_namlen = (short)strlen(dp->d_name);
    dp->d_reclen = (short)(DIRSIZ(dp));
    /*
     * Close the directory
     */
    (void) io_close(chid);

    return dp;
}

#endif

/* FIN */
